(* Verschlsseln / Entschlsseln von Dateien
   optional mit Komprimierung ber ZLib
    J. Rathlev, D-24222 Schwentinental

 -----------------------------------------------------------------------------
   Cipher routines from
     Delphi Encryption Compendium Part I (Version 5.2)
     2006-2008 Hagen Reddmann, Arvid Winkelsdorf
 -----------------------------------------------------------------------------

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - May 2013
         2 - May 2015

   File format:
   ------------
   Byte         Number of bytes
    0           4       Cipher identity
    4           1       Mode + info
    5..8        4       Hash identity
    9           1       Seed length (L)
    10..10+L-1  L       Seed (N = M = 10+L)
      optional: File Info 5 bytes (see below)
      N         1       File attribute
      N+1..N+8  8       Timestamp (TFileTime)
                M = N+9
    M..         Begin of data
    ..          K bytes of data
    M+K         1       Length of Cipher Message Authentication Code (C)
    M+K+1       C       Cipher Message Authentication Code

    MethodInfo:
    -----------
         Bit       7 | 6 | 5 | 4 | 3 | 2 | 1 | 0
                    |  |           |   |   |   |
                    |  |           ----------------> Cipher mode (see DecCipher)
                    |  |           0 = cmCTSx
                    |  |           1 = cmCBCx (default)
                    |  |           2 = cmCFB8
                    |  |           3 = cmCFBx
                    |  |           4 = cmOFB8
                    |  |           5 = cmOFBx
                    |  |           6 = cmCFS8
                    |  |           7 = cCFSx
                    |  |           8 = cmECBx
                    |  ---> Header includes file info (time stamp and attributes)
                    ------> Data are compressed
     *)

unit CipherTools;

interface

uses
  Winapi.Windows, System.Classes, System.SysUtils, DecCipher, DecHash, ZLibex,
  GzUtils;

const
  // Cipher algos
  cmRijndael = 0;  // AES
  cmBlowfish = 1;
  cmTripleDES = 2;
  cmTwofish  = 3;

  cmCompress = $80;  // set bit 7 of cipher algo if a compressed file (gz) is encrypted
  cmFileInfo = $40;  // include a header with fileinfo: time stamp, file attr.
  defCipherMode = cmRijndael;

  CipherAlgos = '"AES 256",Blowfish,Triple-DES,Twofish';

  CryptExt = '.cry';

type
  // base class for crypted processing
  TCryptThread = class (TGZipThread)
  private
    FPwd             : AnsiString;
    FFileMode        : byte;
    FCipherClass     : TDecCipherClass;
    FAttr,
    FHashLength      : integer;
    FFileInfo,
    FCompress        : boolean;
    FFileTime        : TFileTime;
  protected
    procedure ErrorBreak (Mode,Error : integer);
    procedure Execute; override;
  public
    constructor Create  (ASourceName,
                         ADestName      : string;
                         Password       : AnsiString;
                         ASuspend       : Boolean;
                         ABufSize       : integer;
                         APriority      : TThreadPriority);
    end;

  // encrypt files
  TEncryptThread = class (TCryptThread)
  protected
    procedure SetCipherAlgo (ACipherAlgo : integer);
    procedure EncryptStream (sSource,sDest : TStream);
    procedure Execute; override;
  public
    constructor Create  (ASourceName,
                         ADestName      : string;
                         Password       : AnsiString;
                         ACipherAlgo    : integer;
                         ASuspend       : Boolean;
                         ABufSize       : integer;
                         APriority      : TThreadPriority);
    property CipherAlgo : integer write SetCipherAlgo;
    end;

  //decrypt files
  TDecryptThread = class (TCryptThread)
  private
    FCipherName : string;
    function GetCipherAlgo : integer;
  protected
    procedure SetFileMode (Mode : byte);
    function DecryptStream (sSource,sDest : TStream) : integer;
    procedure Execute; override;
  public
    constructor Create  (ASourceName,
                         ADestName      : string;
                         Password       : AnsiString;
                         ASuspend       : Boolean;
                         ABufSize       : integer;
                         APriority      : TThreadPriority);
    property CipherAlgo : integer read GetCipherAlgo;
    end;

implementation

uses WinApiUtils, FileUtils, UnitConsts, DecUtil, FileErrors;

{ ------------------------------------------------------------------- }
constructor TCryptThread.Create (ASourceName,
                                 ADestName      : string;
                                 Password       : AnsiString;
                                 ASuspend       : Boolean;
                                 ABufSize       : integer;
                                 APriority      : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize,APriority);
  FPwd:=Password;
  FCompress:=false; FFileInfo:=false;
  FFileTime:=DateTimeToFileTime(Now); FAttr:=0;
  FErrFlag:=errOK;
  end;

// Mode = 0 : no freeing, 1 = free source, 2 = free dest, 3 free both
procedure TCryptThread.ErrorBreak (Mode,Error : integer);
begin
  FError:=FErrFlag or Error;
  if (Mode and 1)<>0 then try sSource.Free; except end;
  if (Mode and 2)<>0 then try sDest.Free; except end;
  Terminate;
  end;

procedure TCryptThread.Execute;
begin
  Terminate;
  end;

{ ------------------------------------------------------------------- }
constructor TEncryptThread.Create (ASourceName,ADestName : string;
                                   Password              : AnsiString;
                                   ACipherAlgo    : integer;
                                   ASuspend       : Boolean;
                                   ABufSize       : integer;
                                   APriority      : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,Password,ASuspend,ABufSize,APriority);
  CipherAlgo:=ACipherAlgo;
  FErrFlag:=errEncrypt;
  end;

procedure TEnCryptThread.SetCipherAlgo (ACipherAlgo : integer);
begin
  FFileMode:=ACipherAlgo and $F0;
  FCompress:=ACipherAlgo and cmCompress <>0;
  FFileInfo:=ACipherAlgo and cmFileInfo <>0;
  case ACipherAlgo and $F of
  cmBlowfish  : FCipherClass:=TCipher_Blowfish;
  cmTripleDES : FCipherClass:=TCipher_3DES;
  cmTwofish   : FCipherClass:=TCipher_Twofish;
  else FCipherClass:=TCipher_Rijndael;
    end;
  end;

procedure TEncryptThread.EncryptStream (sSource,sDest : TStream);
var
  n,i    : integer;
  b1,b2  : byte;
  Seed   : Binary;

  procedure Write(const Value; Size: Integer);
  begin
    sDest.WriteBuffer(Value, Size);
  end;

  procedure WriteByte(Value: Byte);
  begin
    Write(Value, SizeOf(Value));
  end;

  procedure WriteCardinal(Value: cardinal);
  begin
    Write(Value, SizeOf(Value));
  end;

  procedure WriteInt64(Value: int64);
  begin
    Write(Value, SizeOf(Value));
  end;

  procedure WriteBinary(const Value: Binary);
  begin
    WriteByte(Length(Value));
    Write(Value[1], Length(Value));
  end;

begin
  with FCipherClass.Create do begin
    Mode:=cmCBCx;
    Init(FPwd);
    Seed:=RandomBinary(16);
    try
      Init(THash_SHA1.KDFx(FPwd,Seed,Context.KeySize));
      WriteCardinal(Identity);
      WriteByte(Byte(Mode) or FFileMode);
      WriteCardinal(THash_SHA1.Identity);
      WriteBinary(Seed);
      if FFileInfo then begin
        WriteByte(FAttr);
        WriteInt64(TInt64(FFileTime).AsInt64);
        end;
      WriteInt64(sSource.Size);
      EncodeStream(sSource,sDest,sSource.Size);
      WriteBinary(CalcMAC);
    finally
      Free;
      end;
    end;
  end;

procedure TEncryptThread.Execute;
var
  sSource,sDest     : TFileStream;
  sTmp              : TStream;
  errorcode,n       : Integer;
begin
  if FileExists (SourceName) and (length(DestName)>0) then begin
    FFileTime:=GetFileLastWriteTime(SourceName);
    FAttr:=FileGetAttr(SourceName);
    try
      sSource:=TFileStream.Create(SourceName,fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do begin
        ErrorBreak(0,errFileOpen); Exit;
        end;
      end;
    // Ziel immer berschreiben
    if FileExists(DestName) then begin
      if FileSetAttr(DestName,faArchive)<>0 then begin
        ErrorBreak(1,errFileAttr); Exit;
        end;
      end;
    try
      sDest:=TFileStream.Create(DestName,fmCreate);
    except
      on EFCreateError do begin
        ErrorBreak(1,errFileCreate); Exit;
        end;
      end;
    // prfen, ob mit Kompression
    if FCompress then begin      // see: SetAlgo
      sTmp:=TMemoryStream.Create;
      errorcode:=GzCompress(sSource,sTmp);
      if errorcode=errOK then begin
        sTmp.Position:=0;
        try
          EncryptStream(sTmp,sDest);
          sTmp.Free;
        except
          ErrorBreak(3,errFileWrite); Exit;
          end;
        end
      else begin
        sTmp.Free;
        ErrorBreak(3,errorcode); Exit;
        end;
      end
    else begin
      try
        EncryptStream(sSource,sDest);
      except
        ErrorBreak(3,errFileWrite); Exit;
        end;
      end;
    ProtectStream(sSource);
    try
      sSource.Free;
    except
      on EFileStreamError do begin
        ErrorBreak(2,errFileClose); Exit;
        end;
      end;
    try
      sDest.Free;
    except
      on EFileStreamError do begin
        ErrorBreak(0,errFileClose); Exit;
        end;
      end;
    if not FFileInfo then begin
      n:=SetFileLastWriteTime(DestName,FFileTime);
      if n=0 then n:=FileSetAttr(DestName,FAttr);
      if n>0 then begin
        ErrorBreak(0,errFileAttr); exit;
        end;
      end;
    end;
  Terminate;
  end;

{ ------------------------------------------------------------------- }
constructor TDeCryptThread.Create (ASourceName,
                                   ADestName      : string;
                                   Password       : AnsiString;
                                   ASuspend       : Boolean;
                                   ABufSize       : integer;
                                   APriority      : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,Password,ASuspend,ABufSize,APriority);
  FCipherName:='';
  FErrFlag:=errDecrypt;
  end;

procedure TDecryptThread.SetFileMode (Mode : byte);
begin
  FFileMode:=Mode and $F0;
  FCompress:=Mode and cmCompress <>0;
  FFileInfo:=Mode and cmFileInfo <>0;
  end;

function TDecryptThread.GetCipherAlgo : integer;
begin
  Result:=-1;
  if AnsiSameText(FCipherName,TCipher_Rijndael.ClassName) then Result:=0
  else if AnsiSameText(FCipherName,TCipher_Blowfish.ClassName) then Result:=1
  else if AnsiSameText(FCipherName,TCipher_3DES.ClassName) then Result:=2
  else if AnsiSameText(FCipherName,TCipher_Twofish.ClassName) then Result:=3;
  end;

function TDecryptThread.DecryptStream (sSource,sDest : TStream) : integer;
var
  b : byte;
  sTmp  : TStream;

  procedure Read(var Value; Size: Integer);
  begin
    sSource.ReadBuffer(Value, Size);
  end;

  function ReadByte: Byte;
  begin
    Read(Result, SizeOf(Result));
  end;

  function ReadCardinal: LongWord;
  begin
    Read(Result, SizeOf(Result));
  end;

  function ReadInt64: int64;
  begin
    Read(Result, SizeOf(Result));
  end;

  function ReadBinary: Binary;
  begin
    SetLength(Result, ReadByte);
    Read(Result[1], Length(Result));
  end;

begin
   //  decode
  Result:=errOK;
  try
    with CipherByIdentity(ReadCardinal).Create do
    try
      FCipherName:=ClassName;
      b:=ReadByte;
      Mode:=TCipherMode(b and $F);
      SetFileMode(b);
      Init(HashByIdentity(ReadCardinal).KDFx(FPwd, ReadBinary, Context.KeySize));
      if FFileInfo then begin
        FAttr:=ReadByte;
        TInt64(FFileTime).AsInt64:=ReadInt64;
        end;
      if FCompress then sTmp:=TMemoryStream.Create
      else sTmp:=sDest;
      DecodeStream(sSource,sTmp,ReadInt64);
      if ReadBinary<>CalcMAC then Result:=errFileDCrypt;
    finally
      Free;
      end;
    if FCompress then begin
      if (Result=errOK) then begin
        sTmp.Position:=0;
        Result:=GzUncompress(sTmp,sDest,sTmp.Size);
        end;
      sTmp.Free;
      end;
  except
    ProtectStream(sDest);
    end;
  end;

procedure TDecryptThread.Execute;
var
  sSource,sDest     : TFileStream;
  sTmp              : TStream;
  errorcode,n,i     : Integer;
begin
  if FileExists (SourceName) and (length(DestName)>0) then begin
    FFileTime:=GetFileLastWriteTime(SourceName);
    FAttr:=FileGetAttr(SourceName);
    try
      sSource:=TFileStream.Create(SourceName,fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do begin
        ErrorBreak(0,errFileOpen); Exit;
        end;
      end;
    // Ziel immer berschreiben
    if FileExists(DestName) then begin
      if FileSetAttr(DestName,faArchive)<>0 then begin
        ErrorBreak(1,errFileAttr); Exit;
        end;
      end;
    try
      sDest:=TFileStream.Create(DestName,fmCreate);
    except
      on EFCreateError do begin
        ErrorBreak(1,errFileCreate); Exit;
        end;
      end;
    errorCode:=DecryptStream(sSource,sDest);
    if errorcode<>errOk then begin
      ErrorBreak(2,errorcode); Exit;
      end;
    try
      sSource.Free;
    except
      on EFileStreamError do begin
        ErrorBreak(2,errFileClose); Exit;
        end;
      end;
    try
      sDest.Free;
    except
      on EFileStreamError do begin
        ErrorBreak(0,errFileClose); Exit;
        end;
      end;
    n:=SetFileLastWriteTime(DestName,FFileTime);
    if n=0 then n:=FileSetAttr(DestName,FAttr);
    if n>0 then ErrorBreak(0,errFileAttr);
    end;
  Terminate;
  end;

initialization
  RegisterDECClasses([TCipher_Rijndael, TCipher_Blowfish, TCipher_3DES, TCipher_Twofish, THash_SHA1]);
  IdentityBase := $7A3F42CD;
end.
